/* 
**  mod_myauth.c -- MySQL authentication module for Apache 2
**
**  Copyright (C) Svein Olav Bjerkeset, January 2004
**  License: GNU GENERAL PUBLIC LICENSE Version 2
*/

#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "apr_strings.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include <mysql.h>

#include <config.h>

#define ENC_PLAIN  1
#define ENC_CRYPT  2
#define ENC_MYSQL  4	/* default */

#ifdef USE_OPENSSL
#include <openssl/bio.h>
#include <openssl/evp.h>
#define ENC_MD5BASE64 8
#else
#define ENC_MD5BASE64 0
#endif

typedef struct {
	char		*my_host;
	char		*my_db;
	uint16_t	my_port;
	char		*my_login;
	char		*my_passw;
	int			my_authoritative;
	char		*my_usertbl;
	char		*my_userfld;
	char		*my_pwfld;
	char		*my_grpfld;
	char		*my_grptbl;
	int			my_enctype;
} myauth_config_rec;


static int myauth_init_handler(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s);
static int myauth_auth_checker (request_rec *r);
static int myauth_check_user_id (request_rec *r);


static void *myauth_init_config(apr_pool_t *p, char *d)
{
	myauth_config_rec *cfg = (myauth_config_rec*) 
		apr_pcalloc(p, sizeof(myauth_config_rec));

	cfg->my_port    = 0;
	cfg->my_host    = apr_pstrdup (p, "localhost");
	cfg->my_db      = NULL;
	cfg->my_login   = NULL;
	cfg->my_passw   = NULL;
	cfg->my_usertbl = apr_pstrdup (p, "users");
	cfg->my_userfld = apr_pstrdup (p, "name");
	cfg->my_pwfld   = apr_pstrdup (p, "password");
	cfg->my_grptbl  = apr_pstrdup (p, "groups");
	cfg->my_grpfld  = apr_pstrdup (p, "gname");
	cfg->my_enctype = ENC_MYSQL;
	cfg->my_authoritative = 1;
	return cfg;
}


static const char* myauth_set_slot (cmd_parms *parms, void *mconfig, const char *arg)
{
	char **addr = mconfig + (int)parms->info;

	*addr = apr_pstrdup(parms->pool, arg);
	return NULL;
}


static const char* myauth_set_port(cmd_parms *parms, void *mconfig, const char *arg)
{
	myauth_config_rec *cfg = (myauth_config_rec *)mconfig;
	cfg->my_port = (unsigned) atoi (arg);
	return NULL;
}


static const char* myauth_set_enctype(cmd_parms *parms, void *mconfig, 
	const char *word1, const char *word2)
{
	myauth_config_rec *cfg = (myauth_config_rec *)mconfig;
	const char *arg[] = {word1, word2, NULL};
	int i;
	
	cfg->my_enctype = 0;
	for (i=0; arg[i]; i++) {
		if (strcasecmp (arg[i], "PlainText") == 0)
			cfg->my_enctype |= ENC_PLAIN;
		if (strcasecmp (arg[i], "Crypt") == 0)
			cfg->my_enctype |= ENC_CRYPT;
		if (strcasecmp (arg[i], "MySQL") == 0)
			cfg->my_enctype |= ENC_MYSQL;
		if (strcasecmp (arg[i], "MD5Base64") == 0)
			cfg->my_enctype |= ENC_MD5BASE64;
		if (strcasecmp (arg[i], "All") == 0)
			cfg->my_enctype |= (ENC_PLAIN|ENC_CRYPT|ENC_MYSQL|ENC_MD5BASE64);
	}
	return NULL;
}


static const command_rec myauth_commands[] = {
	AP_INIT_TAKE1( "MyAuthHost", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_host),
		OR_AUTHCFG, "host to connect to." ),
	AP_INIT_TAKE1("MyAuthPort", myauth_set_port,
		NULL,
		OR_AUTHCFG, "port to connect to." ),
	AP_INIT_TAKE1("MyAuthDB", myauth_set_slot, 
		(void *) APR_OFFSETOF(myauth_config_rec, my_db),
		OR_AUTHCFG, "database to use for authentication." ),
	AP_INIT_TAKE1("MyAuthDBUser", myauth_set_slot, 
		(void *) APR_OFFSETOF(myauth_config_rec, my_login),
		OR_AUTHCFG, "username to use for connection." ),
	AP_INIT_TAKE1("MyAuthDBPassw", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_passw),
		OR_AUTHCFG, "password to use for connection." ),
	AP_INIT_TAKE1("MyAuthUserTable", myauth_set_slot,	
		(void *) APR_OFFSETOF(myauth_config_rec, my_usertbl),
		OR_AUTHCFG, "table to use for authentication." ),
	AP_INIT_TAKE1("MyAuthUserField", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_userfld),
		OR_AUTHCFG, "user field in table." ),
	AP_INIT_TAKE1("MyAuthPasswordField", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_pwfld),
		OR_AUTHCFG, "password field in table." ),
	AP_INIT_TAKE1("MyAuthGroupTable", myauth_set_slot,	
		(void *) APR_OFFSETOF(myauth_config_rec, my_grptbl),
		OR_AUTHCFG, "table holding group memberships." ),
	AP_INIT_TAKE1("MyAuthGroupField", myauth_set_slot,	
		(void *) APR_OFFSETOF(myauth_config_rec, my_grpfld),
		OR_AUTHCFG, "name of group name column." ),
	AP_INIT_FLAG("MyAuthAuthoritative",  ap_set_flag_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_authoritative),
		OR_AUTHCFG, "set authoritative on or off." ), 
	AP_INIT_TAKE12("MyAuthEncryption", myauth_set_enctype,
		NULL,
		OR_AUTHCFG, "set encryption methods." ),
	{ NULL }
};


static void myauth_register_hooks(apr_pool_t *p)
{
	ap_hook_post_config(myauth_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_auth_checker(myauth_auth_checker, NULL, NULL, APR_HOOK_FIRST);
	ap_hook_check_user_id(myauth_check_user_id, NULL, NULL, APR_HOOK_FIRST);
}


/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA myauth_module = {
    STANDARD20_MODULE_STUFF, 
    myauth_init_config,    /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    myauth_commands,       /* table of config file commands       */
    myauth_register_hooks  /* register hooks                      */
};


static char *get_pw (request_rec *r, myauth_config_rec *cfg)
{
	MYSQL		*dbh = NULL;
	MYSQL_RES	*res = NULL;
	MYSQL_ROW	row;
	char		*sql, *pw = NULL;

	if (!(dbh = mysql_init(NULL))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: cannot initialize MySQL");
		return NULL;
	}
	
	if (!mysql_real_connect(dbh, cfg->my_host, cfg->my_login, cfg->my_passw,
		cfg->my_db, cfg->my_port, NULL, CLIENT_COMPRESS)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: connection failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	sql = apr_psprintf(r->pool, 
		"SELECT %s FROM %s WHERE %s = '%s'", 
		cfg->my_pwfld, cfg->my_usertbl, cfg->my_userfld, r->user);
	if (mysql_query(dbh, sql)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: query failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	if (!(res = mysql_store_result(dbh))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: could not store query result: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	switch (mysql_num_rows(res)) {
	case 1:
		if (!(row = mysql_fetch_row(res))) {
			ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
				"myauth: could not get row: %s", mysql_error(dbh));
			break;
		}
		pw = apr_pstrdup (r->pool, row[0]);
		break;
	case 0:
		break;
	default:
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: user '%s' not unique", r->user);
		break;
	}
	
	mysql_free_result(res);
	mysql_close(dbh);
	return pw;
}


static apr_table_t *get_groups (request_rec *r, myauth_config_rec *cfg)
{
	MYSQL		*dbh = NULL;
	MYSQL_RES	*res = NULL;
	MYSQL_ROW	row;
	char		*sql, *pw = NULL;
	apr_table_t *grptbl = apr_table_make (r->pool, 15);

	if (!(dbh = mysql_init(NULL))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: cannot initialize MySQL");
		return NULL;
	}
	
	if (!mysql_real_connect(dbh, cfg->my_host, cfg->my_login, cfg->my_passw,
		cfg->my_db, cfg->my_port, NULL, CLIENT_COMPRESS)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: connection failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	sql = apr_psprintf(r->pool, 
		"SELECT %s FROM %s WHERE %s = '%s'", 
		cfg->my_grpfld, cfg->my_grptbl, cfg->my_userfld, r->user);
	if (mysql_query(dbh, sql)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: group query failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	if (!(res = mysql_store_result(dbh))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"myauth: could not store group query result: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	while (row = mysql_fetch_row(res))
		apr_table_setn (grptbl, apr_pstrdup(r->pool, row[0]), "in");
	
	mysql_free_result(res);
	mysql_close(dbh);
	return grptbl;
}

#ifdef USE_OPENSSL
/**
  Database and OS independent encryption: Create an MD5 digest from
  msg, Base64 encode the obtained digest and store the result in buf.

  @param msg	string to encode
  @param buf	where to store the null terminated result of encoding.
  				If buflen is actually smaller than the result, the result
				is truncated, but still null-terminated.
  @param buflen	max. length of buf (including null-terminator)

  @return  the length of the encoded string.
*/
static int md5base64(const char *msg, char *buf, int buflen)
{
    EVP_MD_CTX mdctx;
    const EVP_MD *md;
    unsigned char md_value[EVP_MAX_MD_SIZE];
    int md_len, i;
    BIO *mem, *bio, *b64;

    memset(buf, 0, buflen);

    /* digest md5 */
    md = EVP_md5();
    if(!md) {
        return 0;
    }
    EVP_MD_CTX_init(&mdctx);
    EVP_DigestInit_ex(&mdctx, md, NULL);
    EVP_DigestUpdate(&mdctx, msg, strlen(msg));
    EVP_DigestFinal_ex(&mdctx, md_value, &md_len);
    EVP_MD_CTX_cleanup(&mdctx);

    /* base64 encoding */
    b64 = BIO_new(BIO_f_base64());
    mem = BIO_new(BIO_s_mem());
    bio = BIO_push(b64, mem);

    BIO_write(bio, md_value, md_len);
    BIO_flush(bio);

    i = BIO_read(mem, buf, buflen);
    if (i > 0) {
        i--; /* kill the newline char */
    } else {
       i = 0;
    }
    buf[i] = '\0';
    BIO_free_all(bio);
    return i;
}
#endif

static int valid_pw (request_rec *r, const char *sent, const char *real, const int etyp)
{
	char my_pw[64];

#ifdef USE_OPENSSL
	if (etyp & ENC_MD5BASE64) {
		int i = md5base64(sent, my_pw, sizeof(my_pw));
        if (i > 0 && strcmp(real, my_pw) == 0)
            return 1;
    }
#endif
	if (etyp & ENC_MYSQL) {
#ifdef HAVE_MAKE_SCRAMBLED_PASSWORD
		make_scrambled_password (my_pw, sent);
		if (!strcmp (real, my_pw))
			return 1;
#endif
#ifdef HAVE_MAKE_SCRAMBLED_PASSWORD_323
		make_scrambled_password_323 (my_pw, sent);
		if (!strcmp (real, my_pw))
			return 1;
#endif
	}
	if ((etyp & ENC_CRYPT) && !strcmp (real,crypt(sent,real)))
		return 1;
	if ((etyp & ENC_PLAIN) && !strcmp (real,sent))
		return 1;
	return 0;
}


/* Module initialization handler */
static int myauth_init_handler(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_add_version_component(pconf, "MyAuth/1.4");
	return OK;
}


/* Check for valid user */
static int myauth_auth_checker (request_rec *r)
{
	int i, restricted = 0;
	const apr_array_header_t *reqs_all = ap_requires(r);
	const char *t, *w;
	require_line *reql;
	myauth_config_rec *conf = 
		ap_get_module_config(r->per_dir_config, &myauth_module);
	apr_table_t *groups = get_groups (r, conf);

	/* no require directive => authorize all users */
	if (!reqs_all)
		return OK;
		
	reql = (require_line *)reqs_all->elts;
	for (i=0; i<reqs_all->nelts; i++) {
		
		/* check if this method is masked out by LIMIT directive */
		if (!(reql[i].method_mask & (AP_METHOD_BIT << r->method_number)))
			continue; 
		
		restricted = 1;
		t = reql[i].requirement;
		w = ap_getword_white (r->pool, &t);

		if (!strcmp (w, "valid-user"))
			return OK;
		
		if (!strcmp (w, "user"))
			while (t[0]) {
				w = ap_getword_conf (r->pool, &t);
				if (!strcmp (r->user, w))
					return OK;
			}

		else if (!strcmp (w, "group"))
			while (groups && t[0]) {
				w = ap_getword_conf (r->pool, &t);
				if (apr_table_get (groups, w))
					return OK;
			}
		
		else if (conf->my_authoritative) 
			ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
				"myauth: unknown require directive '%s'", reql[i].requirement);
		
	}
	
	if (!restricted)
		return OK;

	if (!(conf->my_authoritative))
		return DECLINED;
	
	ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
		"myauth: access to '%s' failed: user '%s' not authorized", 
		r->uri, r->user);
	ap_note_basic_auth_failure(r);
	return HTTP_UNAUTHORIZED;
}


/* Authorization check */
static int myauth_check_user_id (request_rec *r)
{
    myauth_config_rec *conf = 
		ap_get_module_config(r->per_dir_config, &myauth_module);
    const char *sent_pw;
    char *real_pw;
    int res;

    if ((res = ap_get_basic_auth_pw(r, &sent_pw))) {
        return res;
    }

    if (!conf->my_db || !conf->my_login)
        return DECLINED;

    if (!(real_pw = get_pw(r, conf))) {
        if (!(conf->my_authoritative))
            return DECLINED;
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
                      "myauth: user '%s' not found: %s", r->user, r->uri);
        ap_note_basic_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    if (!valid_pw (r, sent_pw, real_pw, conf->my_enctype)) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
                      "myauth: user '%s' authentication failure for '%s': "
                      "incorrect password",
                      r->user, r->uri);
        ap_note_basic_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }
    return OK;
}
